#include "system.h"




/* drive geometries */
#define DG144_HEADS       2     /* heads per drive (1.44M) */
#define DG144_TRACKS     80     /* number of tracks (1.44M) */
#define DG144_SPT        18     /* sectors per track (1.44M) */
#define DG144_GAP3FMT  0x54     /* gap3 while formatting (1.44M) */
#define DG144_GAP3RW   0x1b     /* gap3 while reading/writing (1.44M) */

#define DG168_HEADS       2     /* heads per drive (1.68M) */
#define DG168_TRACKS     80     /* number of tracks (1.68M) */
#define DG168_SPT        21     /* sectors per track (1.68M) */
#define DG168_GAP3FMT  0x0c     /* gap3 while formatting (1.68M) */
#define DG168_GAP3RW   0x1c     /* gap3 while reading/writing (1.68M) */

/* IO ports */
#define FDC_DOR  (0x3f2)   /* Digital Output Register */
#define FDC_MSR  (0x3f4)   /* Main Status Register (input) */
#define FDC_DRS  (0x3f4)   /* Data Rate Select Register (output) */
#define FDC_DATA (0x3f5)   /* Data Register */
#define FDC_DIR  (0x3f7)   /* Digital Input Register (input) */
#define FDC_CCR  (0x3f7)   /* Configuration Control Register (output) */

/* command bytes (these are 765 commands + options such as MFM, etc) */
#define CMD_SPECIFY (0x03)  /* specify drive timings */
#define CMD_WRITE   (0xc5)  /* write data (+ MT,MFM) */
#define CMD_READ    (0xe6)  /* read data (+ MT,MFM,SK) */
#define CMD_RECAL   (0x07)  /* recalibrate */
#define CMD_SENSEI  (0x08)  /* sense interrupt status */
#define CMD_FORMAT  (0x4d)  /* format track (+ MFM) */
#define CMD_SEEK    (0x0f)  /* seek track */
#define CMD_VERSION (0x10)  /* FDC version */

static volatile bool done = false;
static bool dchange = false;
static bool motor = false;
static int mtick = 0;
static volatile int tmout = 0;
static uint8 status[7] = { 0 };
static uint8 statsz = 0;
static uint8 sr0 = 0;
static uint8 fdc_track = 0xff;
static DrvGeom geometry = { DG144_HEADS,DG144_TRACKS,DG144_SPT };

static long tbaddr;    /* physical address of track buffer located below 1M */

/* helper functions */

typedef struct DmaChannel {
   uint8 page;     /* page register */
   uint8 offset;   /* offset register */
   uint8 length;   /* length register */
} DmaChannel;

const static DmaChannel dmainfo[] = {
   { 0x87, 0x00, 0x01 },
   { 0x83, 0x02, 0x03 },
   { 0x81, 0x04, 0x05 },
   { 0x82, 0x06, 0x07 }
};


long alloc_dma_buffer(int size)
{
   return 0x1000;
}

/*
 * this sets up a DMA trasfer between a device and memory.  Pass the DMA
 * channel number (0..3), the physical address of the buffer and transfer
 * length.  If 'read' is TRUE, then transfer will be from memory to device,
 * else from the device to memory.
 */
void dma_xfer(int channel,long physaddr,int length,bool read)
{
   long page,offset;
   
   assert(channel < 4);
   
   /* calculate dma page and offset */
   page = physaddr >> 16;
   offset = physaddr & 0xffff;
   length -= 1;  /* with dma, if you want k bytes, you ask for k - 1 */

   _CLI_
   
   /* set the mask bit for the channel */
   outportb(0x0a,channel | 4);
   
   /* clear flipflop */
   outportb(0x0c,0);
   
   /* set DMA mode (write+single+r/w) */
   outportb(0x0b,(read ? 0x48 : 0x44) + channel);
   
   /* set DMA page */
   outportb(dmainfo[channel].page,page);
   
   /* set DMA offset */
   outportb(dmainfo[channel].offset,offset & 0xff);  /* low byte */
   outportb(dmainfo[channel].offset,offset >> 8);    /* high byte */
   
   /* set DMA length */
   outportb(dmainfo[channel].length,length & 0xff);  /* low byte */
   outportb(dmainfo[channel].length,length >> 8);    /* high byte */
   
   /* clear DMA mask bit */
   outportb(0x0a,channel);
   
   _STI_  /* enable irq's */
}

void int1c()
{
   if (tmout) --tmout;     /* bump timeout */
   
   if (mtick > 0)
     --mtick;
   else if (!mtick && motor) {
      outportb(FDC_DOR,0x0c);  /* turn off floppy motor */
      motor = false;
   }
}


/* init driver */
void floppy_install()
{
   int i;
   
   /* allocate track buffer (must be located below 1M) */
   tbaddr = alloc_dma_buffer(512);
   
   /* install IRQ6 handler */
   irq_install_handler(6,floppy_handler);
   System.puts("IRQ6 handler installed\n");

   reset();

   /* get floppy controller version */
   senduint8(CMD_VERSION);
   i = getuint8();

   if (i == 0x80)
     System.puts("NEC765 controller found\n");
   else
     System.puts("enhanced controller found\n");
}

/* deinit driver */
void deinit(void)
{
   /* stop motor forcefully */
   outportb(FDC_DOR,0x0c);
}

/* senduint8() routine from intel manual */
void senduint8(int value)
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inportb(FDC_MSR);
      if ((msr & 0xc0) == 0x80) {
	 outportb(FDC_DATA,value);
	 return;
      }
      inportb(0x80);   /* delay */
   }
}

/* getuint8() routine from intel manual */
int getuint8()
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inportb(FDC_MSR);
      if ((msr & 0xd0) == 0xd0) {
	 return inportb(FDC_DATA);
      }
      inportb(0x80);   /* delay */
   }

   return -1;   /* read timeout */
}

/* this waits for FDC command to complete */
bool waitfdc(bool sensei)
{
   tmout = 100;   /* set timeout to 1 second */
     
   /* wait for IRQ6 handler to signal command finished */
   while (!done && tmout)
     ;
   
   /* read in command result uint8s */
   statsz = 0;
   while ((statsz < 7) && (inportb(FDC_MSR) & (1<<4))) {
      status[statsz++] = getuint8();
   }

   if (sensei) {
      /* send a "sense interrupt status" command */
      senduint8(CMD_SENSEI);
      sr0 = getuint8();
      fdc_track = getuint8();
   }
   
   done = false;
   
   if (!tmout) {
      /* timed out! */
      if (inportb(FDC_DIR) & 0x80)  /* check for diskchange */
	dchange = true;
      
      return false;
   } else
     return true;
}

/* This is the IRQ6 handler */
void floppy_handler(regs*t)
{
   /* signal operation finished */
   done = true;
   System.putc('h');
}


/*
 * converts linear block address to head/track/sector
 * 
 * blocks are numbered 0..heads*tracks*spt-1
 * blocks 0..spt-1 are serviced by head #0
 * blocks spt..spt*2-1 are serviced by head 1
 * 
 * WARNING: garbage in == garbage out
 */
void block2hts(int block,int *head,int *track,int *sector)
{
   *head = (block % (geometry.spt * geometry.heads)) / (geometry.spt);
   *track = block / (geometry.spt * geometry.heads);
   *sector = block % geometry.spt + 1;
}

/**** disk operations ****/

/* this gets the FDC to a known state */
void reset(void)
{
   /* stop the motor and disable IRQ/DMA */
   outportb(FDC_DOR,0);
   
   mtick = 0;
   motor = false;

   /* program data rate (500K/s) */
   outportb(FDC_DRS,0);

   /* re-enable interrupts */
   outportb(FDC_DOR,0x0c);

   /* resetting triggered an interrupt - handle it */
   done = true;
   waitfdc(true);

   /* specify drive timings (got these off the BIOS) */
   senduint8(CMD_SPECIFY);
   senduint8(0xdf);  /* SRT = 3ms, HUT = 240ms */
   senduint8(0x02);  /* HLT = 16ms, ND = 0 */
   
   /* clear "disk change" status */
   seek(1);
   recalibrate();

   dchange = false;
}

/* this returns whether there was a disk change */
bool diskchange(void)
{
   return dchange;
}

/* this turns the motor on */
void motoron(void)
{
   if (!motor) {
      mtick = -1;     /* stop motor kill countdown */
      outportb(FDC_DOR,0x1c);
      timer_wait(50); /* delay 500ms for motor to spin up */
      motor = true;
   }
}

/* this turns the motor off */
void motoroff(void)
{
   if (motor) {
      mtick = 200;   /* start motor kill countdown: 36 ticks ~ 2s */
   }
}

/* recalibrate the drive */
void recalibrate(void)
{
   /* turn the motor on */
   motoron();
   
   /* send actual command uint8s */
   senduint8(CMD_RECAL);
   senduint8(0);

   /* wait until seek finished */
   waitfdc(true);
   
   /* turn the motor off */
   motoroff();
}

/* seek to track */
bool seek(int track)
{
   if (fdc_track == track)  /* already there? */
     return true;
   
   motoron();
   
   /* send actual command uint8s */
   senduint8(CMD_SEEK);
   senduint8(0);
   senduint8(track);

   /* wait until seek finished */
   if (!waitfdc(true))
     return false;     /* timeout! */

   /* now let head settle for 15ms */
   timer_wait(2);
   
   motoroff();
   
   /* check that seek worked */
   if ((sr0 != 0x20) || (fdc_track != track))
     return false;
   else
     return true;
}

/* checks drive geometry - call this after any disk change */
bool log_disk(DrvGeom *g)
{
   /* get drive in a known status before we do anything */
   reset();

   /* assume disk is 1.68M and try and read block #21 on first track */
   geometry.heads = DG168_HEADS;
   geometry.tracks = DG168_TRACKS;
   geometry.spt = DG168_SPT;

   if (read_block(20,0)) {
      /* disk is a 1.68M disk */
      if (g) {
	 g->heads = geometry.heads;
	 g->tracks = geometry.tracks;
	 g->spt = geometry.spt;
      }
      return true;             
   }
   
   /* OK, not 1.68M - try again for 1.44M reading block #18 on first track */
   geometry.heads = DG144_HEADS;
   geometry.tracks = DG144_TRACKS;
   geometry.spt = DG144_SPT;

   if (read_block(17,0)) {
      /* disk is a 1.44M disk */
      if (g) {
	 g->heads = geometry.heads;
	 g->tracks = geometry.tracks;
	 g->spt = geometry.spt;
      }
      return true;
   }
   
   /* it's not 1.44M or 1.68M - we don't support it */
   return false;
}

/* read block (blockbuff is 512 uint8 buffer) */
bool read_block(int block,uint8 *blockbuff)
{
   return fdc_rw(block,blockbuff,true);
}

/* write block (blockbuff is 512 uint8 buffer) */
bool write_block(int block,uint8 *blockbuff)
{
   return fdc_rw(block,blockbuff,false);
}

/*
 * since reads and writes differ only by a few lines, this handles both.  This
 * function is called by read_block() and write_block()
 */
bool fdc_rw(int block,uint8 *blockbuff,bool read)
{
   int head,track,sector,tries,i;

   /* convert logical address into physical address */
   block2hts(block,&head,&track,&sector);
//   System.puts("block %d = %d:%02d:%02d\n",block,head,track,sector);
   
   /* spin up the disk */
   motoron();

   if (!read && blockbuff) {
      /* copy data from data buffer into track buffer */
      memcpy((void*)tbaddr,blockbuff,512);

   }
   
   for (tries = 0;tries < 3;tries++) {
      /* check for diskchange */
      if (inportb(FDC_DIR) & 0x80) {
	 dchange = true;
	 seek(1);  /* clear "disk change" status */
	 recalibrate();
	 motoroff();
	 return false;
      }
         
      /* move head to right track */
      if (!seek(track)) {
	 motoroff();
	 return false;
      }
      
      /* program data rate (500K/s) */
      outportb(FDC_CCR,0);
      
      /* send command */
      if (read) {
	 dma_xfer(2,tbaddr,512,false);
	 senduint8(CMD_READ);
      } else {
	 dma_xfer(2,tbaddr,512,true);
	 senduint8(CMD_WRITE);
      }
      
      senduint8(head << 2);
      senduint8(track);
      senduint8(head);
      senduint8(sector);
      senduint8(2);               /* 512 uint8s/sector */
      senduint8(geometry.spt);
      if (geometry.spt == DG144_SPT)
	senduint8(DG144_GAP3RW);  /* gap 3 size for 1.44M read/write */
      else
	senduint8(DG168_GAP3RW);  /* gap 3 size for 1.68M read/write */
      senduint8(0xff);            /* DTL = unused */
      
      /* wait for command completion */
      /* read/write don't need "sense interrupt status" */
      if (!waitfdc(false))
	return false;   /* timed out! */
      
      if ((status[0] & 0xc0) == 0) break;   /* worked! outta here! */
   
      recalibrate();  /* oops, try again... */
   }
   
   /* stop the motor */
   motoroff();

   if (read && blockbuff) {

      memcpy((void*)blockbuff,(void*)tbaddr,512);

   }

   System.puts("status uint8s: ");
   for (i = 0;i < statsz;i++)
     System.puti(status[i],16);

   System.puts("\n");

   return (tries != 3);
}

/* this formats a track, given a certain geometry */
bool format_track(uint8 track,DrvGeom *g)
{
   int i,h,r,r_id,split;
   uint8 tmpbuff[256];

   /* check geometry */
   if (g->spt != DG144_SPT && g->spt != DG168_SPT)
     return false;
   
   /* spin up the disk */
   motoron();

   /* program data rate (500K/s) */
   outportb(FDC_CCR,0);

   seek(track);  /* seek to track */

   /* precalc some constants for interleave calculation */
   split = g->spt / 2;
   if (g->spt & 1) split++;
   
   for (h = 0;h < g->heads;h++) {
      /* for each head... */
      
      /* check for diskchange */
      if (inportb(FDC_DIR) & 0x80) {
	 dchange = true;
	 seek(1);  /* clear "disk change" status */
	 recalibrate();
	 motoroff();
	 return false;
      }

      i = 0;   /* reset buffer index */
      for (r = 0;r < g->spt;r++) {
	 /* for each sector... */

	 /* calculate 1:2 interleave (seems optimal in my system) */
	 r_id = r / 2 + 1;
	 if (r & 1) r_id += split;
	 
	 /* add some head skew (2 sectors should be enough) */
	 if (h & 1) {
	    r_id -= 2;
	    if (r_id < 1) r_id += g->spt;
	 }
      
	 /* add some track skew (1/2 a revolution) */
	 if (track & 1) {
	    r_id -= g->spt / 2;
	    if (r_id < 1) r_id += g->spt;
	 }
	 
	 /**** interleave now calculated - sector ID is stored in r_id ****/

	 /* fill in sector ID's */
	 tmpbuff[i++] = track;
	 tmpbuff[i++] = h;
	 tmpbuff[i++] = r_id;
	 tmpbuff[i++] = 2;
      }

      /* copy sector ID's to track buffer */
memcpy((void*)tbaddr,(void*)tmpbuff,i);
      
      /* start dma xfer */
      dma_xfer(2,tbaddr,i,true);
      
      /* prepare "format track" command */
      senduint8(CMD_FORMAT);
      senduint8(h << 2);
      senduint8(2);
      senduint8(g->spt);
      if (g->spt == DG144_SPT)      
	senduint8(DG144_GAP3FMT);    /* gap3 size for 1.44M format */
      else
	senduint8(DG168_GAP3FMT);    /* gap3 size for 1.68M format */
      senduint8(0);     /* filler uint8 */
	 
      /* wait for command to finish */
      if (!waitfdc(false))
	return false;
      
      if (status[0] & 0xc0) {
	 motoroff();
	 return false;
      }
   }
   
   motoroff();
   
   return true;
}

